@@ -8,18 +8,18 @@ Author: Gabriel Ebner
 #pragma once
 #include "library/messages.h"
 #include "kernel/environment.h"
-#include "util/json.hpp"
+#include "util/lean_json.h"
 
 namespace lean {
 
-using json = nlohmann::json;
-
 json json_of_severity(message_severity sev);
 
 json json_of_message(message const & msg);
 
 json json_of_name(name const &);
 
+json to_json(name const &);
+
 void add_source_info(environment const & env, name const & d, json & record);
 json serialize_decl(name const & short_name, name const & long_name, environment const & env, options const & o);
 json serialize_decl(name const & d, environment const & env, options const & o);
