@@ -20,6 +20,7 @@ Author: Leonardo de Moura
 #include "library/io_state.h"
 #include "library/io_state_stream.h"
 #include "library/message_builder.h"
+#include "library/tactic/tactic_log.h"
 #include "library/tactic/tactic_state.h"
 #include "frontends/lean/parser_state.h"
 #include "frontends/lean/local_decls.h"
@@ -35,6 +36,8 @@ class metavar_context;
 class local_context_adapter;
 class scope_message_context;
 
+#define AST_TOP_ID 1
+
 class parser_info {
 protected:
     environment             m_env;
@@ -53,9 +56,7 @@ protected:
     // profiling
     bool                   m_profile;
 
-    // If the following flag is true we do not raise error messages
-    // noncomputable definitions not tagged as noncomputable.
-    bool                   m_ignore_noncomputable;
+    noncomputable_policy    m_noncomputable_policy;
 
     // error recovery
     bool                   m_error_recovery = true;
@@ -125,10 +126,10 @@ protected:
     parser_scope mk_parser_scope(optional<options> const & opts = optional<options>());
     void restore_parser_scope(parser_scope const & s);
 
-    virtual std::tuple<expr, expr, name> parse_definition(buffer<name> & lp_names, buffer<expr> & params,
+    virtual std::tuple<expr, expr, name> parse_definition(ast_data * parent, buffer<name> & lp_names, buffer<expr> & params,
                                                           bool is_example, bool is_instance, bool, bool) = 0;
 
-    friend environment single_definition_cmd_core(parser_info & p, decl_cmd_kind kind, cmd_meta meta);
+    friend environment single_definition_cmd_core(parser_info & p, decl_cmd_kind kind, ast_data * parent, cmd_meta meta);
 public:
     void updt_options();
     options get_options() const { return m_ios.get_options(); }
@@ -139,16 +140,19 @@ public:
     virtual optional<pos_info> get_pos_info(expr const & e) const = 0;
     virtual message_builder mk_message(pos_info const &p, message_severity severity) const = 0;
     virtual message_builder mk_message(message_severity severity) const = 0;
-    bool ignore_noncomputable() const { return m_ignore_noncomputable; }
-    void set_ignore_noncomputable() { m_ignore_noncomputable = true; }
+    noncomputable_policy get_noncomputable_policy() const { return m_noncomputable_policy; }
+    void set_noncomputable_policy(noncomputable_policy p) { m_noncomputable_policy = p; }
     virtual char const * get_file_name() const = 0;
     virtual pos_info cmd_pos() const = 0;
     virtual optional<pos_info> const & get_break_at_pos() const = 0;
     virtual parser_pos_provider get_parser_pos_provider(pos_info const & some_pos) const = 0;
+    virtual std::shared_ptr<tactic_log> get_tactic_log() { return nullptr; }
     expr mk_sorry(pos_info const & p, bool synthetic = false);
     bool has_error_recovery() const { return m_error_recovery; }
 };
 
+typedef rb_map<unsigned, ast_id, unsigned_cmp> tag_ast_table;
+
 class parser : public abstract_parser, public parser_info {
     name_generator          m_ngen;
     bool                    m_use_exceptions;
@@ -156,12 +160,18 @@ class parser : public abstract_parser, public parser_info {
     std::string             m_file_name;
     scanner                 m_scanner;
     token_kind              m_curr;
-    bool                    m_imports_parsed;
+    bool                    m_imports_parsed = false;
     bool                    m_scanner_inited = false;
     parser_scope_stack      m_quote_stack;
     pos_info                m_last_cmd_pos;
     unsigned                m_next_tag_idx;
+    ast_id                  m_vm_parser_ast_id = 0;
     pos_info_table          m_pos_table;
+    tag_ast_table           m_tag_ast_table;
+    std::vector<ast_data*>  m_ast;
+    bool                    m_ast_invalid = false;
+    ast_id                  m_commands = 0;
+    std::shared_ptr<tactic_log> m_tactic_log;
     // By default, when the parser finds a unknown identifier, it signs an error.
     // When the following flag is true, it creates a constant.
     // This flag is when we are trying to parse mutually recursive declarations.
@@ -175,22 +185,19 @@ class parser : public abstract_parser, public parser_info {
     pos_info               m_last_recovered_error_pos {0, 0};
     optional<pos_info>     m_backtracking_pos;
 
-    // curr command token
-    name                   m_cmd_token;
-
     void sync_command();
 
     tag get_tag(expr e);
 
     unsigned curr_level_lbp() const;
-    level parse_max_imax(bool is_max);
-    level parse_level_id();
-    level parse_level_nud();
-    level parse_level_led(level left);
+    pair<ast_id, level> parse_max_imax(bool is_max);
+    pair<ast_id, level> parse_level_id();
+    pair<ast_id, level> parse_level_nud();
+    pair<ast_id, level> parse_level_led(pair<ast_id, level> left);
 
-    void parse_mod_doc_block();
+    ast_id parse_mod_doc_block();
 
-    void process_imports();
+    void process_imports(ast_data * parent);
     void process_postponed(buffer<expr> const & args, bool is_left, buffer<notation::action_kind> const & kinds,
                            buffer<list<expr>> const & nargs, buffer<expr> const & ps, buffer<pair<unsigned, pos_info>> const & scoped_info,
                            list<notation::action> const & postponed, pos_info const & p, buffer<expr> & new_args);
@@ -199,18 +206,18 @@ class parser : public abstract_parser, public parser_info {
     expr parse_led_notation(expr left);
     expr parse_inaccessible();
     expr parse_placeholder();
-    expr parse_anonymous_var_pattern();
+    // expr parse_anonymous_var_pattern();
     expr parse_nud();
     bool curr_starts_expr();
     expr parse_numeral_expr(bool user_notation = true);
     expr parse_decimal_expr();
     expr parse_string_expr();
     expr parse_char_expr();
-    expr parse_inst_implicit_decl();
-    void parse_inst_implicit_decl(buffer<expr> & r);
+    pair<ast_id, expr> parse_inst_implicit_decl();
+    void parse_inst_implicit_decl(ast_data * parent, buffer<expr> & r);
     expr parse_binder_core(binder_info const & bi, unsigned rbp);
-    bool parse_binder_collection(buffer<pair<pos_info, name>> const & names, binder_info const & bi, buffer<expr> & r);
-    void parse_binder_block(buffer<expr> & r, binder_info const & bi, unsigned rbp, bool allow_default);
+    ast_id parse_binder_collection(buffer<pair<pos_info, name>> const & names, binder_info const & bi, buffer<expr> & r);
+    ast_id parse_binder_block(buffer<expr> & r, binder_info const & bi, unsigned rbp, bool allow_default);
     struct parse_binders_config {
         /* (input) If m_allow_empty is true, then parse_binders will succeed even if not binder is parsed. */
         bool     m_allow_empty{false};
@@ -259,26 +266,27 @@ class parser : public abstract_parser, public parser_info {
 
           The default is `RelaxedImplicit` */
         implicit_infer_kind * m_infer_kind{nullptr};
+        /* (output) The AST ID associated to the m_infer_kind */
+        ast_id * m_infer_kind_id{nullptr};
         /* (output) It is set to true if the last binder is surrounded
            with some kind of bracket (e.g., '()', '{}', '[]'). */
         bool     m_last_block_delimited{false};
         /* (output) If m_nentries != nullptr, then local notation declarations are stored here */
         buffer<notation_entry> * m_nentries{nullptr};
     };
-    void parse_binders_core(buffer<expr> & r, parse_binders_config & cfg);
-    local_environment parse_binders(buffer<expr> & r, parse_binders_config & cfg);
-    bool parse_local_notation_decl(buffer<notation_entry> * entries);
+    void parse_binders_core(ast_data * parent, buffer<expr> & r, parse_binders_config & cfg);
+    local_environment parse_binders(ast_data * parent, buffer<expr> & r, parse_binders_config & cfg);
+    ast_id parse_local_notation_decl(buffer<notation_entry> * entries);
 
-    pair<optional<name>, expr> parse_id_tk_expr(name const & tk, unsigned rbp);
+    std::tuple<ast_id, optional<name>, expr> parse_id_tk_expr(name const & tk, unsigned rbp);
 
-    friend environment section_cmd(parser & p);
-    friend environment namespace_cmd(parser & p);
-    friend environment end_scoped_cmd(parser & p);
+    friend environment section_cmd(parser & p, ast_id &);
+    friend environment namespace_cmd(parser & p, ast_id &);
+    friend environment end_scoped_cmd(parser & p, ast_id &);
 
     std::shared_ptr<snapshot> mk_snapshot();
 
-    optional<expr> resolve_local(name const & id, pos_info const & p, list<name> const & extra_locals,
-                                 bool allow_field_notation = true);
+    optional<expr> resolve_local(name const & id, pos_info const & p, list<name> const & extra_locals);
 
     friend class module_parser;
     friend class patexpr_to_expr_fn;
@@ -296,10 +304,30 @@ public:
            module_loader const & import_fn,
            std::istream & strm, std::string const & file_name,
            bool use_exceptions = false);
+    parser(parser const &) = delete;
     ~parser();
 
     void init_scanner();
 
+    ast_data & new_ast(name type, pos_info start, name value = {});
+    void finalize_ast(ast_id id, expr const & e);
+    bool is_ast_invalid() { return m_ast_invalid; }
+    void set_ast_expr(ast_id id, expr e);
+    ast_data & get_ast(ast_id id) { return *m_ast[id]; }
+    ast_id get_id(expr const & e) const;
+    ast_data & expr_ast(expr const & e) { return get_ast(get_id(e)); }
+    const char * ast_string(ast_id id) const { return m_ast[id]->m_value.get_string(); }
+    pos_info ast_pos(ast_id id) const { return m_ast[id]->m_start; }
+    void push_from_vm_parse(ast_id id) { if (m_vm_parser_ast_id) get_ast(m_vm_parser_ast_id).push(id); }
+    ast_id get_vm_parse_parent() { return m_vm_parser_ast_id; }
+    ast_id set_vm_parse_parent(ast_id id) {
+        auto old = m_vm_parser_ast_id;
+        m_vm_parser_ast_id = id;
+        return old;
+    }
+    ast_data & new_modifiers(cmd_meta & meta);
+    friend void export_ast(parser const &);
+
     void from_snapshot(snapshot const & snap);
 
     name next_name() { return m_ngen.next(); }
@@ -343,7 +371,8 @@ public:
     local_expr_decls const & get_local_expr_decls() const { return m_local_decls; }
 
     /** \brief Return the current position information */
-    virtual pos_info pos() const override final { return pos_info(m_scanner.get_line(), m_scanner.get_pos()); }
+    virtual pos_info pos() const override final { return m_scanner.get_pos_info(); }
+    pos_info end_pos() const { return m_scanner.get_last_end_pos_info(); }
     virtual expr save_pos(expr const & e, pos_info p) override final;
     expr rec_save_pos(expr const & e, pos_info p) override final;
     expr rec_save_pos(expr const & e, optional<pos_info> p) override final { return p ? rec_save_pos(e, *p) : e; }
@@ -352,12 +381,13 @@ public:
     pos_info pos_of(expr const & e, pos_info default_pos) const;
     pos_info pos_of(expr const & e) const { return pos_of(e, pos()); }
     pos_info cmd_pos() const override { return m_last_cmd_pos; }
-    name const & get_cmd_token() const { return m_cmd_token; }
 
     parser_pos_provider get_parser_pos_provider(pos_info const & some_pos) const override {
         return parser_pos_provider(m_pos_table, m_file_name, some_pos, m_next_tag_idx);
     }
 
+    std::shared_ptr<tactic_log> get_tactic_log() override;
+
     expr mk_app(expr fn, expr arg, pos_info const & p);
     expr mk_app(expr fn, buffer<expr> const & args, pos_info const & p);
     expr mk_app(std::initializer_list<expr> const & args, pos_info const & p);
@@ -404,81 +434,83 @@ public:
     void check_token_or_id_next(name const & tk, char const * msg);
     /** \brief Check if the current token is an identifier, if it is return it and move to next token,
         otherwise throw an exception. */
-    name check_id_next(char const * msg, break_at_pos_exception::token_context ctxt =
+    pair<ast_id, name> check_id_next(char const * msg, break_at_pos_exception::token_context ctxt =
             break_at_pos_exception::token_context::none);
     void check_not_internal(name const & n, pos_info const & pos);
     /** \brief Similar to check_id_next, but also ensures the identifier is *not* an internal/reserved name. */
-    name check_decl_id_next(char const * msg, break_at_pos_exception::token_context ctxt =
+    pair<ast_id, name> check_decl_id_next(char const * msg, break_at_pos_exception::token_context ctxt =
             break_at_pos_exception::token_context::none);
     /** \brief Check if the current token is an atomic identifier, if it is, return it and move to next token,
         otherwise throw an exception. */
-    name check_atomic_id_next(char const * msg);
-    name check_atomic_decl_id_next(char const * msg);
+    pair<ast_id, name> check_atomic_id_next(char const * msg);
+    pair<ast_id, name> check_atomic_decl_id_next(char const * msg);
     list<name> to_constants(name const & id, char const * msg, pos_info const & p) const;
     name to_constant(name const & id, char const * msg, pos_info const & p);
     /** \brief Check if the current token is a constant, if it is, return it and move to next token, otherwise throw an exception. */
-    name check_constant_next(char const * msg);
+    pair<ast_id, name> check_constant_next(char const * msg);
 
     mpq const & get_num_val() const { return m_scanner.get_num_val(); }
     name const & get_name_val() const { return m_scanner.get_name_val(); }
     std::string const & get_str_val() const { return m_scanner.get_str_val(); }
     token_info const & get_token_info() const { return m_scanner.get_token_info(); }
     std::string const & get_stream_name() const { return m_scanner.get_stream_name(); }
+    std::vector<ast_comment> const & get_comments() const { return m_scanner.get_comments(); }
 
     unsigned get_small_nat();
-    virtual unsigned parse_small_nat() override final;
-    virtual std::string parse_string_lit() override final;
-    double parse_double();
+    virtual pair<ast_id, unsigned> parse_small_nat() override final;
+    virtual pair<ast_id, std::string> parse_string_lit() override final;
+    pair<ast_id, double> parse_double();
 
     /** \brief Parses a documentation block (`/-- TEXT -/`). For example, `/-- Doc\ndoc -/` returns " Doc\ndoc ". */
-    std::string parse_doc_block();
+    pair<ast_id, std::string> parse_doc_block();
 
-    bool parse_local_notation_decl() { return parse_local_notation_decl(nullptr); }
+    ast_id parse_local_notation_decl() { return parse_local_notation_decl(nullptr); }
 
-    level parse_level(unsigned rbp = 0);
+    pair<ast_id, level> parse_level(unsigned rbp = 0);
 
     expr parse_binder(unsigned rbp);
 
-    local_environment parse_binders(buffer<expr> & r, bool & last_block_delimited) {
+    local_environment parse_binders(ast_data * parent, buffer<expr> & r, bool & last_block_delimited) {
         parse_binders_config cfg;
-        auto new_env = parse_binders(r, cfg);
+        auto new_env = parse_binders(parent, r, cfg);
         last_block_delimited = cfg.m_last_block_delimited;
         return new_env;
     }
 
-    local_environment parse_binders(buffer<expr> & r, unsigned rbp, bool allow_default = false) {
+    local_environment parse_binders(ast_data * parent, buffer<expr> & r, unsigned rbp, bool allow_default = false) {
         parse_binders_config cfg;
         cfg.m_rbp           = rbp;
         cfg.m_allow_default = allow_default;
-        return parse_binders(r, cfg);
+        return parse_binders(parent, r, cfg);
     }
 
-    void parse_simple_binders(buffer<expr> & r, unsigned rbp) {
+    void parse_simple_binders(ast_data * parent, buffer<expr> & r, unsigned rbp) {
         parse_binders_config cfg;
         cfg.m_simple_only = true;
         cfg.m_rbp         = rbp;
-        parse_binders(r, cfg);
+        parse_binders(parent, r, cfg);
     }
 
-    local_environment parse_optional_binders(buffer<expr> & r, bool allow_default = false, bool explicit_delimiters = false) {
+    local_environment parse_optional_binders(ast_data * parent, buffer<expr> & r, bool allow_default = false, bool explicit_delimiters = false) {
         parse_binders_config cfg;
         cfg.m_allow_empty         = true;
         cfg.m_allow_default       = allow_default;
         cfg.m_explicit_delimiters = explicit_delimiters;
-        return parse_binders(r, cfg);
+        return parse_binders(parent, r, cfg);
     }
 
-    local_environment parse_optional_binders(buffer<expr> & r, implicit_infer_kind & kind) {
+    local_environment parse_optional_binders(ast_data * parent, buffer<expr> & r, ast_id & kind_id, implicit_infer_kind & kind) {
         parse_binders_config cfg;
         cfg.m_allow_empty   = true;
         cfg.m_infer_kind    = &kind;
-        return parse_binders(r, cfg);
+        cfg.m_infer_kind_id = &kind_id;
+        return parse_binders(parent, r, cfg);
     }
 
-    local_environment parse_binders(buffer<expr> & r, buffer<notation_entry> & nentries) {
+    local_environment parse_binders(ast_data * parent, buffer<expr> & r, buffer<notation_entry> & nentries) {
         parse_binders_config cfg;
         cfg.m_nentries = &nentries;
-        return parse_binders(r, cfg);
+        return parse_binders(parent, r, cfg);
     }
 
     optional<binder_info> parse_optional_binder_info(bool simple_only = false);
@@ -488,7 +520,8 @@ public:
     void parse_close_binder_info(binder_info const & bi) { return parse_close_binder_info(optional<binder_info>(bi)); }
 
     /** \brief Convert an identifier into an expression (constant or local constant) based on the current scope */
-    expr id_to_expr(name const & id, pos_info const & p, bool resolve_only = false, bool allow_field_notation = true,
+    expr id_to_expr(name const & id, ast_data & data,
+                    bool resolve_only = false, bool allow_field_notation = true,
                     list<name> const & extra_locals = list<name>());
 
     /** Always parses an expression.  Returns a synthetic sorry even if no input is consumed. */
@@ -501,10 +534,10 @@ public:
     /** \brief Parse an (optionally) qualified expression.
         If the input is of the form <id> : <expr>, then return the pair (some(id), expr).
         Otherwise, parse the next expression and return (none, expr). */
-    pair<optional<name>, expr> parse_qualified_expr(unsigned rbp = 0);
+    std::tuple<ast_id, optional<name>, expr> parse_qualified_expr(unsigned rbp = 0);
     /** \brief If the input is of the form <id> := <expr>, then return the pair (some(id), expr).
         Otherwise, parse the next expression and return (none, expr). */
-    pair<optional<name>, expr> parse_optional_assignment(unsigned rbp = 0);
+    std::tuple<ast_id, optional<name>, expr> parse_optional_assignment(unsigned rbp = 0);
 
     /** \brief Parse a pattern or expression */
     expr parse_pattern_or_expr(unsigned rbp = 0);
@@ -537,9 +570,9 @@ public:
     expr parse_scoped_expr(buffer<expr> const & ps, unsigned rbp = 0) { return parse_scoped_expr(ps.size(), ps.data(), rbp); }
     expr parse_expr_with_env(local_environment const & lenv, unsigned rbp = 0);
 
-    bool parse_command_like();
-    void parse_command(cmd_meta const & meta);
-    bool parse_imports(unsigned & fingerprint, std::vector<module_name> &);
+    bool parse_command_like(ast_data * parent = nullptr);
+    ast_id parse_command(cmd_meta const & meta);
+    bool parse_imports(unsigned & fingerprint, ast_data * parent, std::vector<module_name> &);
     bool imports_parsed() const { return m_imports_parsed; }
 
     struct quote_scope {
@@ -555,7 +588,7 @@ public:
     bool in_quote() const { return m_in_quote; }
 
     void maybe_throw_error(parser_error && err) override;
-    level parser_error_or_level(parser_error && err);
+    pair<ast_id, level> parser_error_or_level(parser_error && err);
     expr parser_error_or_expr(parser_error && err);
     void throw_invalid_open_binder(pos_info const & pos);
 
@@ -610,7 +643,7 @@ public:
     virtual char const * get_file_name() const override;
 
 protected:
-    virtual std::tuple<expr, expr, name> parse_definition(buffer<name> & lp_names, buffer<expr> & params,
+    virtual std::tuple<expr, expr, name> parse_definition(ast_data * parent, buffer<name> & lp_names, buffer<expr> & params,
                                                           bool is_example, bool is_instance, bool, bool) override;
 };
 
@@ -618,7 +651,6 @@ bool parse_commands(environment & env, io_state & ios, char const * fname);
 
 class dummy_def_parser : public parser_info {
 public:
-  bool            m_ignore_noncomputable;
   pos_info        m_pos;
   std::string     m_file_name;
   name            m_name;
@@ -633,7 +665,6 @@ public:
     parser_info(env, io_state(io_state(), opts))
       { m_error_recovery = false; }
   char const * get_file_name() const override { return m_file_name.c_str(); }
-  bool ignore_noncomputable() const { return m_ignore_noncomputable; }
   options get_options() const { return m_ios.get_options(); }
   message_builder mk_message(pos_info const &p, message_severity severity) const override {
     std::shared_ptr<abstract_type_context> tc = std::make_shared<type_context_old>(m_env, get_options());
@@ -662,7 +693,7 @@ public:
   optional<expr> get_well_founded_tac() const { return m_wf_tac; }
 
 protected:
-  virtual std::tuple<expr, expr, name> parse_definition(buffer<name> & lp_names, buffer<expr> & params,
+  virtual std::tuple<expr, expr, name> parse_definition(ast_data * parent, buffer<name> & lp_names, buffer<expr> & params,
                                                         bool is_example, bool is_instance, bool, bool) override;
 };
 
